#!/usr/bin/env python

########################################################################################################################
#  S M S   C L A S S                                                                                                   #
#  Tato trida je urcena pro komunikaci s GSM modemem Siemens GSM modem TC35 apod.                                      #
#  Umi cist, posilat, mazat SMS a prozvanet.                                                                           #
#  Soucasti je i trida PDU, ktera dovede enkodovat a dekodovat SMS zpravy z/do PDU formatu                             #
########################################################################################################################




# Import variuos libraries
import sys, serial, re, time, string
from time import gmtime, strftime, localtime


########################################################################################################################
# Trida implementuje komunikacni metody pro GSM modem co se SMS tyce                                                   #
########################################################################################################################

class sms:

    # atribut pro instanci tridy serial
    port = None

    # 
    logprefix = ""
    loglist = []


    # ------------------------------------------------------------------------------------------------------------------
    # Konstruktor
    # Vstup: cislo nebo nazev portu
    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, port):
	self.port = serial.Serial()
	self.port.port = port
	self.port.baudrate = 9600
	self.port.bytesize = 8
	self.port.parity = 'N'
	self.port.stopbits = 1
	self.port.timeout = 10
	self.port.xonxoff = 0
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Privatni metoda na cteni odpovedi od modemu s vyhodnocovanim
    # Vstup:  vzor    - regex vzor pro true vystup
    #         timeout - cas na cteni dalsiho radku z portu
    # Vystup: True/False dle vyhodnoceni vzoru v prijatem radku
    # ------------------------------------------------------------------------------------------------------------------
    def __ReadOneAnswer(self, vzor, timeout):
	self.port.timeout = timeout
	ret = False
	while True:
	    x = self.port.readline()
	    if x=="":
		break
	    self.__ToLog('RX',repr(x))
            if re.match(vzor, x):
        	ret = True
        	break
    	return ret
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Privatni metody pro logovani
    # Princip je ten, aby po ukonceni nejake verejne metody zustal v atributu loglist seznam radku s logem prubehu
    # ------------------------------------------------------------------------------------------------------------------
    def __ToLog(self, prefix, text):
	self.loglist.append(strftime("%Y.%m.%d %H:%M:%S ", localtime()) + '[SMSCLASS]['+self.logprefix+']['+prefix+'] '+text)

    def __ClearAndSetLog(self, prefix):
	self.loglist = []
	self.logprefix = prefix
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Privatni metoda na odeslani prikazu modemu
    # Vstup:  text - retezec, ktery bude odeslan na port
    # ------------------------------------------------------------------------------------------------------------------
    def __Send(self, text):
	self.__ToLog('TX', repr(text))
	self.port.flushInput()
	self.port.write(text)
    # ------------------------------------------------------------------------------------------------------------------



    # ------------------------------------------------------------------------------------------------------------------
    # Metoda na pripojeni k modemu a nastaveni pracovnich pameti
    # Vystup: True/False dle toho, zda-li se podarilo pripojit k portu a detekovat zivy modem
    # ------------------------------------------------------------------------------------------------------------------
    def Connect(self):
	self.__ClearAndSetLog('CONNECT')
	self.__ToLog('', 'Zkousim otevrit port')
	self.port.open()
	if self.port.isOpen()==0:
	    self.__ToLog('ERR', 'Nemohu otevrit port -> fatalni chyba')
	    return False
	# dotaz na dostupnost
	self.__ToLog('OK', 'Port byl otevren, testuji dostupnost modemu prikazem AT')
	self.__Send('AT\r')
	if not self.__ReadOneAnswer('^OK',2):
	    self.__ToLog('ERR', 'Modem neodpovida')
	    return False
	self.__ToLog('OK', 'Modem je dostupny, nastavuji pameti pro ukladani SMS')
	self.__Send('AT+CPMS="MT","MT","MT"\r')
	self.__ReadOneAnswer('^OK',10)
	self.__ToLog('OK', 'Pripojeno')
	return True
    # ------------------------------------------------------------------------------------------------------------------



    # ------------------------------------------------------------------------------------------------------------------
    # Metoda na smazani SMS
    # Vstup:   id - identifikator zpravy v GSM modemu
    # Vystup:  True/False dle toho, zda-li operace mazani probehla v poradku ci nikoliv
    # ------------------------------------------------------------------------------------------------------------------
    def DeleteSMS(self, id):
	self.__ClearAndSetLog('DELETESMS')
	self.__Send('AT+CMGD=%s\r' % str(id))
	self.port.timeout = 2
	if self.__ReadOneAnswer('^OK',3):
	    self.__ToLog('OK', 'Zprava smazana')
	    return True
	else:
	    log += "ERROR"
	    self.__ToLog('ERR', 'Chyba pri mazani zpravy')
	    return False
    # ------------------------------------------------------------------------------------------------------------------



    # ------------------------------------------------------------------------------------------------------------------
    # Metoda na ziskani seznamu SMS
    # Vystup: seznam hash poli s prijatymi SMS zpravami
    # ------------------------------------------------------------------------------------------------------------------
    def GetListOfSMS(self):
	self.__ClearAndSetLog('GETLISTOFSMS')
	msgs = []
	opdu = PDU()
	self.__Send('AT+CMGL=4\r')
	self.port.timeout = 2
	while True:
	    x = self.port.readline()
	    self.__ToLog('RX', repr(x))
	    if (x=="") or (re.match('^OK', x)):
		break
	    try:
		x = x.strip()
		if re.match("^\+CMGL.*", x):
		    id = int(x.split(':')[1].split(',')[0])
		if len(x)>40:
		    t = opdu.decodeSMS(x)
		    t['id'] = id
		    self.__ToLog('SMS', str(t))
		    msgs.append(t)
	    except Exception:
		self.__ToLog('ERR', 'Chybny format PDU')
	return msgs
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Metoda na odeslani SMS
    # Vstup:  dest - cislo prijemce v mezinarovnim formatu (pr. 420608337726)
    #         text - text SMS zpravy (max. 160 znaku)
    # Vystup: True/False dle prinaku, zda-li byla SMS zprava odeslana ci nikoliv
    # ------------------------------------------------------------------------------------------------------------------
    def SendSMS(self, dest, text):
	self.__ClearAndSetLog('SENDSMS')
	opdu = PDU()
	try:
	    sms = opdu.encodeSMS(dest, text)
	    if len(sms)!=2:
		raise
	except Exception:
	    self.__ToLog('ERR', 'Chyba prevodu do PDU')
	    return False
	self.__ToLog('PDU', str(sms))
	self.__Send('AT+CMGS=%s\r' % str(sms['velikost']))
	if not self.__ReadOneAnswer('^>',2):
	    self.__ToLog('ERR', 'Chyba prikazu AT+CMGS')
	    return False
	self.__Send("%s\x1a" % sms['pdu'])
	if self.__ReadOneAnswer('.*OK.*',10):
	    self.__ToLog('OK', 'SMS odeslana')
	    return True
	else:
	    self.__ToLog('ERR', 'SMS nebyla odeslana')
	    return False
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Metoda na prozvoneni
    # Vstup:  dest - cislo prijemce v mezinarovnim formatu (pr. 420608337726)
    #         cas - doba prozvaneni v sekundach (zde je nutne pocitat s dobou navazani spojeni)
    # ------------------------------------------------------------------------------------------------------------------
    def Ring(self, dest, cas):
	self.__ClearAndSetLog('RING')
	self.__Send('ATD+%s;\r' % dest)
	time.sleep(cas)
	self.__Send('ATH\r')
    # ------------------------------------------------------------------------------------------------------------------












########################################################################################################################
# Trida implementuje kodovani a dekodovani PDU formatu SMS zprav                                                       #
########################################################################################################################

class PDU:


    # Hex to dec conversion array
    __hex2dec = {'0':0,'1':1,'2':2,'3':3,'4':4,'5':5,'6':6,'7':7,'8':8,'9':9,'A':10,'B':11,'C':12,'D':13,'E':14,'F':15 }


    # Konstruktor ------------------------------------------------------------------------------------------------------
    def __init__(self):
	None
    # ------------------------------------------------------------------------------------------------------------------


    # ------------------------------------------------------------------------------------------------------------------
    # Prevede hex na dec
    # ------------------------------------------------------------------------------------------------------------------
    def __hex2int(self,n):
	c1 = n[0]
	c2 = n[1]
	c3 = (self.__hex2dec[c1] * 16) + (self.__hex2dec[c2])
	return int("%s" % c3)

    # ------------------------------------------------------------------------------------------------------------------
    # Prevede dec na hex
    # ------------------------------------------------------------------------------------------------------------------
    def __int2hex(self,n):
	hex = ""
	q = n
	while q > 0:
	    r = q % 16
	    if r == 10: hex = 'A' + hex
	    elif r == 11: hex = 'B' + hex
	    elif r == 12: hex = 'C' + hex
	    elif r == 13: hex = 'D' + hex
	    elif r == 14: hex = 'E' + hex
	    elif r == 15: hex = 'F' + hex
	    else:
		hex = str(r) + hex
	    q = int(q/16)
	if len(hex) % 2 == 1: hex = '0' + hex
	return hex

    # ------------------------------------------------------------------------------------------------------------------
    # Prohodi horni a spodni pulbyte
    # ------------------------------------------------------------------------------------------------------------------
    def __byteSwap(self,byte):
	return "%c%c" % (byte[1], byte[0])

    # ------------------------------------------------------------------------------------------------------------------
    # Precte datum a cas z PDU zpravy
    # ------------------------------------------------------------------------------------------------------------------
    def __parseTimeStamp(self,time):
	y =    self.__byteSwap(time[0:2])
	m =    self.__byteSwap(time[2:4])
	d =    self.__byteSwap(time[4:6])
	hour = self.__byteSwap(time[6:8])
	min =  self.__byteSwap(time[8:10])
	sec =  self.__byteSwap(time[10:12])

	# korekce roku
	if int(y) < 70:
	    y = "20" + y

	return "%s-%s-%s %s:%s:%s" % (y, m, d, hour, min, sec)

    # ------------------------------------------------------------------------------------------------------------------
    # Dekoduje 7mi bitove kodovani zpravy
    # ------------------------------------------------------------------------------------------------------------------
    def __decodeText7Bit(self, src):

	bits = ''

	i = 0
	l = len(src) - 1

	# First, get the bit stream, concatenating all binary represented chars
	while i < l:
		bits += self.__char2bits(src[i:i+2])
		i += 2

	# Now decode those pseudo-8bit octets
	char_nr = 0
	i = 1

	tmp_out = ''
	acumul = ''
	decoded = ''
	while char_nr <= len(bits):
		byte = bits[char_nr + i:char_nr + 8]
		tmp_out += byte + "+" + acumul + " "
		byte += acumul
		c = chr(self.__bits2int(byte))

		decoded += c

		acumul = bits[char_nr:char_nr + i]

		i += 1
		char_nr += 8

		if i==8:
			i = 1
			char_nr
			decoded += chr(self.__bits2int(acumul))
			acumul=''
			tmp_out += "\n"

	return decoded.strip("\x00")

    # ------------------------------------------------------------------------------------------------------------------
    # Konvertuje text na 7-mi bitove kodovani
    # ------------------------------------------------------------------------------------------------------------------
    def __encodeText7Bit(self,src):
	result = []
	count = 0
	last = 0
	for c in src:
		this = ord(c) << (8 - count)
		if count:
			result.append('%02X' % ((last >> 8) | (this & 0xFF)))
		count = (count + 1) % 8
		last = this
	result.append('%02X' % (last >> 8))
	return ''.join(result)

    # ------------------------------------------------------------------------------------------------------------------
    # Znak na bin
    # ------------------------------------------------------------------------------------------------------------------
    def __char2bits(self,char):

	inputChar = self.__hex2int(char)
	mask = 1
	output = ''
	bitNo = 1

	while bitNo <= 8:
		if inputChar & mask > 0:
			output = '1' + output
		else:
			output = '0' + output
		mask = mask<<1
		bitNo += 1

	return output

    # ------------------------------------------------------------------------------------------------------------------
    # Bin na int
    # ------------------------------------------------------------------------------------------------------------------
    def __bits2int(self,bits):
	mask = 1
	i = 0
	end = len(bits) - 1

	result = 0
	while i <= end:
		if bits[end - i] == "1":
			result += mask
		mask = mask << 1
		i += 1

	return result





    # ------------------------------------------------------------------------------------------------------------------
    # Dekodovani PDU zpravy
    # Vstup:  sms - prijata PDU zprava
    # Vystup: pole  odesilatel - cislo odesilatele zpravy v mezinarodnim formatu
    #               cas - datum a cas doruceni
    #               zprava - citelna forma zpravy
    # ------------------------------------------------------------------------------------------------------------------
    def decodeSMS(self, sms):

	self.SMS = {}
	
	# Zjistime pocet oktetu SMSC informace
	smsc_len = self.__hex2int(sms[0:2])
	
	# Presun na pozici PDU flagu
	i = (smsc_len * 2) + 2
	
	# Presun na pozici s delkou cisla odesilatele
	i += 2
	sender_len = self.__hex2int(sms[i:i+2])

	# Presun na pozici cisla odesilatele
	i += 4
	
	# Preskladani cisla odesilatele
	tmp_sender='';
	if sender_len % 2 == 1:
	    sender_len += 1;
	for n in range(i, i+sender_len, 2):
	    tmp_sender += sms[n+1] + sms[n]
	self.SMS['sender'] = tmp_sender.strip('F')
	
	# Presun na pozici datumu a casu, skrz cislo odesilatele a 2 bytes flagu
	i += sender_len + 4
        
	# Get the sending timestamp
	self.SMS['time'] = self.__parseTimeStamp(sms[i:i+14])

	# Presun na pozici delky textu
	i += 14
	mlength = self.__hex2int(sms[i:i+2])

	# Dekodovani zpravy
	if mlength > 0:
	    message_enc = sms[i+2:len(sms)]
	    zprava = self.__decodeText7Bit(message_enc)
	else:
	    zprava = ""
	self.SMS['sms'] = zprava
	
        return(self.SMS)




    # ------------------------------------------------------------------------------------------------------------------
    # Vytvoreni PDU zpravy
    # Vstup:  dest_number - telefonni cislo v mezinarodnim formatu, napr. 420608337726
    #         text - text SMS zpravy, max. 160 znaku, bez diakritiky
    # Vystup: pole velikost   - velikost PDU retezce pro prikaz AT+CMGS
    #         pdu  - PDU retezec
    # ------------------------------------------------------------------------------------------------------------------
    def encodeSMS(self, dest_number, text):

	self.SMS = {}

	# Uvodni byte + hlavicka
	result = "000100"

	# Prida delku cisla prijemce a format cisla na mezinarodni
	result += self.__int2hex(len(dest_number)) + '91'

	# Prida cislo prijemce
	tmp_number = dest_number
	if len(tmp_number) % 2 == 1: tmp_number += 'F'
	i = 0
	while i < len(tmp_number):
	    result = result + self.__byteSwap(tmp_number[i:i+2])
	    i += 2

	# Prida dalsi blbosti
	result += '0000'

	# Prida delku zpravy a text zpravy
	result += self.__int2hex(len(text))
	result += self.__encodeText7Bit(text)
	
	# Vrati pole dvojice velikost a PDU retezec
	self.SMS['pdu'] = result
	self.SMS['velikost'] = (len(result)-2) / 2
        return(self.SMS)




